'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (_ref) {
  var t = _ref.types;

  var cache = {};
  var modelPaths = {};

  function getDeclarator(declarations, identifierName) {
    for (var i = 0, l = declarations.length; i < l; i += 1) {
      var d = declarations[i];
      if (t.isIdentifier(d.id) && d.id.name === identifierName) {
        return d;
      }
    }
    return null;
  }

  function getBinding(_scope, identifierName, withFinalScope) {
    var scope = _scope;
    var binding = scope.bindings[identifierName];
    while (!binding && scope) {
      scope = scope.parent;
      binding = scope.bindings[identifierName];
    }

    return withFinalScope ? {
      binding: binding,
      scope: scope
    } : binding;
  }

  function isRequire(node) {
    return t.isCallExpression(node) && t.isIdentifier(node.callee) && node.callee.name === 'require';
  }

  function getImportRequirePath(scope, identifierName) {
    // identifierName: 'createApp', such as: `import createApp from 'mickey';`
    if (scope.hasBinding(identifierName)) {
      var binding = getBinding(scope, identifierName);
      if (binding) {
        var parent = binding.path.parent;
        if (t.isImportDeclaration(parent)) {
          return parent.source.value;
        } else if (t.isVariableDeclaration(parent)) {
          var declarator = getDeclarator(parent.declarations, identifierName);
          if (declarator && isRequire(declarator.init)) {
            return declarator.init.arguments[0].value;
          }
        }
      }
    }
    return null;
  }

  function isMickeyCallExpression(node, scope) {
    return t.isCallExpression(node) && t.isIdentifier(node.callee) && getImportRequirePath(scope, node.callee.name) === 'mickey';
  }

  function isMickeyInstance(_scope, identifierName) {
    // identifierName: 'app', such as: `const app = createApp({});`
    if (_scope.hasBinding(identifierName)) {
      var _getBinding = getBinding(_scope, identifierName, true),
          binding = _getBinding.binding,
          scope = _getBinding.scope;

      if (binding) {
        var parent = binding.path.parent;
        if (t.isVariableDeclaration(parent)) {
          var declarator = getDeclarator(parent.declarations, identifierName);
          if (declarator && isMickeyCallExpression(declarator.init, scope)) {
            return true;
          }
        }
      }
    }
    return false;
  }

  function isMethodCall(node, scope, methodName) {
    if (!t.isMemberExpression(node)) {
      return false;
    }

    var object = node.object;
    var property = node.property;

    return t.isIdentifier(property) && t.isIdentifier(object) && property.name === methodName && isMickeyInstance(scope, object.name);
  }

  function getRequirePath(node, scope) {
    // eslint-disable-line
    switch (node.type) {
      case 'CallExpression':
        if (t.isLiteral(node.arguments[0])) {
          return node.arguments[0].value;
        }
        break;
      case 'Identifier':
        {
          var path = getImportRequirePath(scope, node.name);
          if (path) {
            return path;
          }
          break;
        }
      default:
        break;
    }

    return null;
  }

  function parseComponentPaths(component, scope) {
    var paths = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : [];
    var openingElement = component.openingElement,
        children = component.children;
    var node = openingElement.name;

    if (node && t.isJSXIdentifier(node) && node.name) {
      var identifierName = node.name;
      if (scope.hasBinding(identifierName)) {
        var binding = getBinding(scope, identifierName);
        if (binding) {
          var componentPath = void 0;
          var parent = binding.path.parent;
          if (t.isImportDeclaration(parent)) {
            componentPath = parent.source.value;
          } else if (t.isVariableDeclaration(parent)) {
            var declarator = getDeclarator(parent.declarations, identifierName);
            if (declarator && isRequire(declarator.init)) {
              componentPath = declarator.init.arguments[0].value;
            }
          }

          // collect relative path
          if (componentPath && componentPath[0] === '.') {
            paths.push({
              name: node.name,
              path: componentPath
            });

            // remove import/require statement of the component
            binding.path.parentPath.remove();
          }
        }
      }
    }

    children.forEach(function (child) {
      return parseComponentPaths(child, scope, paths);
    });

    return paths;
  }

  return {
    visitor: {
      Program: {
        enter: function enter(path) {
          var filename = path.hub.file.opts.filename;

          delete cache[filename];
        }
      },
      CallExpression: function CallExpression(path, _ref2) {
        var opts = _ref2.opts;
        var filename = path.hub.file.opts.filename;

        if (cache[filename]) return;

        var _path$node = path.node,
            callee = _path$node.callee,
            args = _path$node.arguments;

        if (isMethodCall(callee, path.scope, 'model')) {
          var modelPath = getRequirePath(args[0], path.scope);
          if (modelPath) {
            modelPaths[filename] = modelPaths[filename] || [];
            modelPaths[filename].push(modelPath);
          }
          console.log(modelPaths);
        } else if (isMethodCall(callee, path.scope, 'render')) {
          var paths = parseComponentPaths(args[0], path.scope);
          if (paths.length) {
            cache[filename] = true;
            if (!opts.quiet) {
              console.info('[babel-plugin-mickey-hmr] Parsed component paths:');
              paths.forEach(function (item) {
                console.log(' - ' + item.path);
              });
              console.log();
            }

            path.parentPath.replaceWithSourceString((0, _getHmrString2.default)(callee.object.name, paths, modelPaths[filename], args, !opts.disableModel));
          } else if (!opts.quiet) {
            console.log(_chalk2.default.yellow('[babel-plugin-mickey-hmr] Can not parse any component path in "' + filename + '"'));
            console.log();
          }
        }
      }
    }
  };
};

var _chalk = require('chalk');

var _chalk2 = _interopRequireDefault(_chalk);

var _getHmrString = require('./getHmrString');

var _getHmrString2 = _interopRequireDefault(_getHmrString);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

module.exports = exports['default']; /* eslint-disable no-console */